#ifndef M6502_H
#define M6502_H

// See cpu.cc for copyrights.

#include "sizes.h"
#include <endian.h>

enum InterruptType        /* Loop6502() returns:        */
{ INT_NONE=0,             /* No interrupt required      */
  INT_IRQ,                /* Standard IRQ interrupt     */
  INT_NMI,                /* Non-maskable interrupt     */
  INT_QUIT                /* Exit the emulation         */
};
                               /* 6502 status flags:         */
#define	C_FLAG	  0x01         /* 1: Carry occured           */
#define	Z_FLAG	  0x02         /* 1: Result is zero          */
#define	I_FLAG	  0x04         /* 1: Interrupts disabled     */
#define	D_FLAG	  0x08         /* 1: Decimal mode            */
#define	B_FLAG	  0x10         /* Break [0 on stk after int] */
#define	R_FLAG	  0x20         /* Always 1                   */
#define	V_FLAG	  0x40         /* 1: Overflow occured        */
#define	N_FLAG	  0x80         /* 1: Result is negative      */

/** Structured Datatypes *************************************/
/** NOTICE: #define LSB_FIRST for machines where least      **/
/**         signifcant byte goes first.                     **/
/*************************************************************/
typedef union
{
#if __BYTE_ORDER==__LITTLE_ENDIAN
  struct { BYTE l,h; } B;
#else
  struct { BYTE h,l; } B;
#endif
  WORD W;
} PAIR;

#ifdef __GNUC__
#define CONSTFUN __attribute__((const))
#else
#define CONSTFUN
#endif

/** Rd6502()/Wr6502/Op6502() *********************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access. Op6502 is the same **/
/** as Rd6502, but used to read *opcodes* only, when many   **/
/** checks can be skipped to make it fast. It is only       **/
/** required if there is a #define FAST_RDOP.               **/
/************************************ TO BE WRITTEN BY USER **/

void Wr6502(register fWORD Addr, register fBYTE Value);
BYTE Rd6502(register fWORD Addr);
BYTE Op6502(register fWORD Addr) CONSTFUN;

extern BYTE Clocks[];
class m6502cpu
{
public:
    BYTE A,P,X,Y,S;         /* CPU registers and program counter   */
    PAIR PC;

    int IPeriod;            /* Set IPeriod to number of CPU cycles */
    int ICount;             /* between calls to Loop6502()         */
    InterruptType IRequest; /* Set to the INT_IRQ when pending IRQ */
private:
    bool AfterCLI;          /* Private, don't touch                */
    int IBackup;            /* Private, don't touch                */
    
    inline BYTE fetchop()
    {
    	register BYTE i= Op6502(PC.W++);
    	ICount -= Clocks[i];
    	return i;
    }
    inline WORD M_LDWORD()
    {
		register WORD Rg = Op6502(PC.W++);
		Rg |= Op6502(PC.W++) << 8;
		return Rg;
    }
    
    /** Addressing Methods ***************************************/
    /** These macros calculate and return effective addresses.  **/
    /*************************************************************/
    inline WORD MC_Ab()	{ return M_LDWORD(); }
    inline WORD MC_Zp()	{ return BYTE(Op6502(PC.W++)); }
    inline WORD MC_Zx()	{ return BYTE(Op6502(PC.W++)+X); }
    inline WORD MC_Zy()	{ return BYTE(Op6502(PC.W++)+Y); }
    inline WORD MC_Ax()	{ return M_LDWORD()+X; }
    inline WORD MC_Ay()	{ return M_LDWORD()+Y; }
    inline WORD MC_Ix() { register WORD K = BYTE(Op6502(PC.W++) + X); return Op6502(K) | (Op6502(K+1) << 8); }
    inline WORD MC_Iy() { register WORD K = Op6502(PC.W++); return (Op6502(K) | (Op6502(K+1) << 8)) + Y; }
    
    /** Reading From Memory **************************************/
    /** These macros calculate address and read from it.        **/
    /*************************************************************/
    inline BYTE MR_Ab()	{ return Rd6502(MC_Ab()); }
    inline BYTE MR_Im()	{ return Op6502(PC.W++); }
    inline BYTE MR_Zp()	{ return Rd6502(MC_Zp()); }
    inline BYTE MR_Zx()	{ return Rd6502(MC_Zx()); }
    inline BYTE MR_Zy()	{ return Rd6502(MC_Zy()); }
    inline BYTE MR_Ax()	{ return Rd6502(MC_Ax()); }
    inline BYTE MR_Ay()	{ return Rd6502(MC_Ay()); }
    inline BYTE MR_Ix()	{ return Rd6502(MC_Ix()); }
    inline BYTE MR_Iy()	{ return Rd6502(MC_Iy()); }
    
public:
	m6502cpu(int ipe = 1) : IPeriod(ipe)
	{
		// Do not reset here, as the memory
		// has probably not yet been initialized.
	}
    /** Reset6502() **********************************************/
    /** This function can be used to reset the registers before **/
    /** starting execution with Run6502(). It sets registers to **/
    /** their initial values.                                   **/
    /*************************************************************/
	void Reset()
	{
        A = X = Y = 0x00;
        P = Z_FLAG|R_FLAG;
        S = 0xFF;
        PC.B.l = Op6502(0xFFFC);
        PC.B.h = Op6502(0xFFFD);
        ICount = IPeriod;
        IRequest = INT_NONE;
        AfterCLI = false;
	}
    /** Run6502() ************************************************/
    /** This function will run 6502 code until Loop6502() call  **/
    /** returns INT_QUIT. It will return the PC at which        **/
    /** emulation stopped, and current register values in R.    **/
    /*************************************************************/
    WORD Run();

    /** Int6502() ************************************************/
    /** This function will generate interrupt of a given type.  **/
    /** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
    /** will cause a normal interrupt, unless I_FLAG set in R.  **/
    /*************************************************************/
    void Int(InterruptType Type);

    /** Loop6502() ***********************************************/
    /** 6502 emulation calls this function periodically to      **/
    /** check if the system hardware requires any interrupts.   **/
    /** This function must return one of following values:      **/
    /** INT_NONE, INT_IRQ, INT_NMI, or INT_QUIT to exit the     **/
    /** emulation loop.                                         **/
    /************************************ TO BE WRITTEN BY USER **/
    InterruptType Loop();
};

#endif
