#include <cstdio>
#include <cstring> // memset
#include <map>

#include "cpu.hh"
#include "ppu.hh"
#include "papu.hh"
#include "clock.h"

#include "mappers.hh" // FIRST INCLUSION
static map<unsigned, mapper *> mappers;

class bisqnes
{
private:
	friend mapper;
	
	m6502cpu cpu;
	
    bisqppu ppu;
    bisqpapu papu;
	
	BYTE ram[0x800];   // Mirrored four times
	BYTE sram[0x2000]; // Battery RAM
	
	BYTE *rom;
	BYTE rom16count;
	BYTE ctrlbyte;
	BYTE vrom8count;
	BYTE mappernum;
	
	BYTE *rompage[8]; // 4k pages (0x1000)
	
	mapper *mapp;
	
public:
	bisqnes() : cpu(1), rom(NULL), mapp(NULL)
	{
	}
	~bisqnes()
	{
		if(rom)delete[] rom;
	}
	
	void load(const char *fn)
	{
		FILE *fp = fopen(fn, "rb");
		if(!fp)return;

		if(fgetc(fp)=='N'
		&& fgetc(fp)=='E'
		&& fgetc(fp)=='S'
		&& fgetc(fp)=='\32')
		{
			rom16count = fgetc(fp);
			vrom8count = fgetc(fp);
			ctrlbyte = fgetc(fp);
			mappernum = fgetc(fp) | (ctrlbyte>>4);
			fseek(fp, 0x10, SEEK_SET);
			
			if(mappernum >= 0x40)
				mappernum &= 15;
			
			fprintf(stderr, "mappernum %u, 4krom=%u, 4kvrom=%u\n", mappernum, rom16count*4, vrom8count*2);
			map<unsigned, mapper *>::iterator i;
			i = mappers.find(mappernum);
			mapp = (i==mappers.end()) ? NULL : i->second;
			
			if(mapp)
			{
				mapp->vrom4kcount = vrom8count * 2;
				mapp->rom4kcount = rom16count * 4;
			}			
			if(rom)delete[] rom;
			rom = new BYTE[rom16count * 16384];
			fread(&rom[0], 16384, rom16count, fp);
			
			ppu.resizevrom(vrom8count * 8192);
			fread(ppu.vromptr(), 8192, vrom8count, fp);
		}
		
		fclose(fp);
		
		// ROM pages ok, cpu reset etc.
		poweron();
	}
	
	void run()
	{
		cpu.Run();
	}
	
	void poweron()
	{
		unsigned rom4count = rom16count * 4;
    	for(unsigned a=0; a<4; ++a)
    	{
    		rompage[a]   = &rom[             a  * 0x1000];
    		rompage[a+4] = &rom[(rom4count-4+a) * 0x1000];
    	}
		memset(ram, 0, sizeof ram);
		memset(sram, 0, sizeof sram);
		
		if(mapp)mapp->reset();
		cpu.Reset();
	}
public:
    inline void Wr(register fWORD Addr, register fBYTE Value)
    {
    	if(Addr < 0x2000) { ram[Addr & 0x7FF] = Value; return; }
    	if(Addr < 0x6000)
    	{
    		if(Addr == 0x4014)
    		{
    			/* DMA transfer source address */
    			Addr = Value * 0x100;
    			/* This _should_ move one byte per 6 tick()s,
    			 * but it's hard to write.
    			 */
    			for(unsigned a=0; a<256; ++a)
    				ppu.spr_ram[a] = Rd(Addr++);
    			return;
    		}
    		if(Addr == 0x4016)
    		{
    		//	fprintf(stderr, "4016: wrote %02X\n", Value);
    		}
    		if(ppu.write(Addr, Value))return;
    		if(papu.write(Addr, Value))return;
    	}
    	else if(Addr < 0x8000)
    	{
    		if(ctrlbyte&8)
    			sram[Addr & 0x1FFF] = Value;
    	}
    	else
    	{
    		if(mapp)
    			mapp->write(Addr, Value);
    		else
    			fprintf(stderr, "Unknown mapper %u wrote %02X to %04ZX\n", mappernum, Value, Addr);
    	}
    }
    inline BYTE Rd(register fWORD Addr)
    {
    	if(Addr < 0x2000)return ram[Addr & 0x7FF];
    	if(Addr < 0x6000)
    	{
    		BYTE value = 0;
    		if(ppu.read(Addr, value))return value;
    		if(papu.read(Addr, value))return value;
    		//if(Addr == 0x4016){fprintf(stderr, "4016: returning something\n");return 0x41;}
    	}
    	if(Addr < 0x8000)return sram[Addr & 0x1FFF];
    	return rompage[(Addr-0x8000)/0x1000][Addr & 0xFFF];
    }
    inline BYTE Op(register fWORD Addr) const
    {
    	if(Addr < 0x2000)return ram[Addr & 0x7FF];
    	if(Addr < 0x6000)return 0;
    	if(Addr < 0x8000)return sram[Addr & 0x1FFF];
    	Addr -= 0x8000;
    	return rompage[Addr/0x1000][Addr & 0xFFF];
    }
    InterruptType tick()
    {
    	// NTSC PPU has 21.47727 MHz clock.
    	// NTSC CPU has 1.7897725 MHz clock.
    	// The ratio between these is exactly 12.
    	
    	// PAL PPU has 26.601712MHz MHz clock.
    	// PAL CPU has 1.7734474    MHz clock.
    	// The ratio is exactly 15.
    	
    	// NTDec PPU has 20.0000 MHz clock.
    	// NTDec CPU has unknown clock.
    	
    	for(unsigned a=0; a<Clock->CPUclockDivider; ++a)
    	{
	    	InterruptType tmp = ppu.tick();
    		if(tmp != INT_NONE)return tmp;
    	}
    	
    	InterruptType tmp = papu.tick();
   		if(tmp != INT_NONE)return tmp;
    	
    	return INT_IRQ;
    }
};

static bisqnes nes;

#include "mappers.cc"

void Wr6502(register fWORD Addr, register fBYTE Value)
{
	nes.Wr(Addr, Value);
}
BYTE Rd6502(register fWORD Addr)
{
	return nes.Rd(Addr);
}
BYTE Op6502(register fWORD Addr)
{
	return nes.Op(Addr);
}

InterruptType m6502cpu::Loop()
{
	return nes.tick();
}

int main(int argc, const char *const *argv)
{
	nes.load(argc > 1 ? argv[1] : "mariobrs.nes");
	nes.run();
}
